@@ -74,6 +74,7 @@ class PerturbFlow(nn.Module):
                  use_cuda: bool = True,
                  seed: int = 42,
                  zero_bias: bool|list = True,
+                 enumrate: bool = False,
                  dtype = torch.float32, # type: ignore
                  ):
         super().__init__()
@@ -102,6 +103,7 @@ class PerturbFlow(nn.Module):
         else:
             self.use_bias = [not zero_bias] * self.cell_factor_size
         #self.use_bias = not zero_bias
+        self.enumrate = enumrate
         
         self.codebook_weights = None
 
@@ -429,7 +431,12 @@ class PerturbFlow(nn.Module):
                 zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
 
             if self.cell_factor_size>0:
-                zus = self._total_effects(zns, us)
+                if self.enumrate:
+                    idx = torch.argmax(ns, dim=1)
+                    zn_loc = acs_loc[idx]
+                    zus = self._total_effects(zn_loc, us)
+                else:
+                    zus = self._total_effects(zns, us)
                 zs = zns+zus
             else:
                 zs = zns
@@ -631,7 +638,12 @@ class PerturbFlow(nn.Module):
                 #        zus = self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
                 #    else:
                 #        zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
-                zus = self._total_effects(zns, us)
+                if self.enumrate:
+                    idx = torch.argmax(ns, dim=1)
+                    zn_loc = acs_loc[idx]
+                    zus = self._total_effects(zn_loc, us)
+                else:
+                    zus = self._total_effects(zns, us)
                 zs = zns+zus
             else:
                 zs = zns
@@ -681,6 +693,9 @@ class PerturbFlow(nn.Module):
                 zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
         return zus
                         
+    def _get_codebook_identity(self):
+        return torch.eye(self.code_size, **self.options)
+    
     def _get_codebook(self):
         I = torch.eye(self.code_size, **self.options)
         if self.latent_dist=='studentt':
@@ -807,6 +822,36 @@ class PerturbFlow(nn.Module):
         A = np.concatenate(A)
         return A
     
+    def predict(self, xs, us, perturbs_predict:list, perturbs_reference:list, library_sizes=None):
+        perturbs_reference = np.array(perturbs_reference)
+        
+        # basal embedding
+        zs = self.get_basal_embedding(xs)
+        for pert in perturbs_predict:
+            pert_idx = int(np.where(perturbs_reference==pert)[0])
+            us_i = us[:,pert_idx].reshape(-1,1)
+            
+            # factor effect of xs 
+            dzs0 = self.get_cell_response(xs, factor_idx=pert_idx, perturb=us_i)
+            
+            # perturbation effect 
+            ps = np.ones_like(us_i)
+            dzs = self.get_cell_response(xs, factor_idx=pert_idx, perturb=ps)
+            
+            zs = zs + dzs0 + dzs
+        
+        if library_sizes is None:
+            library_sizes = np.sum(xs, axis=1, keepdims=True)
+        elif type(library_sizes) == list:
+            library_sizes = np.array(library_sizes)
+            library_sizes = library_sizes.reshape(-1,1)
+        elif len(library_sizes.shape)==1:
+            library_sizes = library_sizes.reshape(-1,1)
+            
+        counts = self.get_counts(zs, library_sizes=library_sizes)
+        
+        return counts, zs
+    
     def _cell_response(self, xs, factor_idx, perturb):
         #zns,_ = self.encoder_zn(xs)    
         zns,_ = self._get_basal_embedding(xs)   
